@@ -1,28 +1,34 @@
 import asyncio
 import json
+import warnings
 import logging
 import os
+import copy
 from types import LambdaType
 from typing import Any, Dict, List, Optional, Text, Tuple
 
 import numpy as np
 import time
+import re
+from text2digits import text2digits
+t2d = text2digits.Text2Digits()
 
-import opentracing
-
+from metrics.constants import ACTION_PREDICT_EXECUTE_TIME
 from rasa.core import jobs
-from rasa.core.actions.action import Action
-from rasa.core.actions.action import (
-    ACTION_LISTEN_NAME,
-    ActionExecutionRejection,
-    UTTER_PREFIX,
-)
+from rasa.core.actions.action import Action, ACTION_DEFAULT_FALLBACK_NAME
+from rasa.core.actions.action import ACTION_LISTEN_NAME, ActionExecutionRejection
 from rasa.core.channels.channel import (
     CollectingOutputChannel,
     UserMessage,
     OutputChannel,
 )
-from rasa.core.constants import ACTION_NAME_SENDER_ID_CONNECTOR_STR, USER_INTENT_RESTART
+from rasa.core.constants import (
+    ACTION_NAME_SENDER_ID_CONNECTOR_STR,
+    USER_INTENT_RESTART,
+    UTTER_PREFIX,
+    USER_INTENT_BACK,
+    USER_INTENT_OUT_OF_SCOPE,
+)
 from rasa.core.domain import Domain
 from rasa.core.events import (
     ActionExecuted,
@@ -40,6 +46,7 @@ from rasa.core.interpreter import (
     RegexInterpreter,
 )
 from rasa.core.nlg import NaturalLanguageGenerator
+from rasa.core.policies.asr_fallback import AsrFallbackPolicy
 from rasa.core.policies.ensemble import PolicyEnsemble
 from rasa.core.policies.form_nlu_fallback import FormFallbackPolicy
 from rasa.core.tracker_store import TrackerStore
@@ -53,7 +60,7 @@ logger = logging.getLogger(__name__)
 MAX_NUMBER_OF_PREDICTIONS = int(os.environ.get("MAX_NUMBER_OF_PREDICTIONS", "10"))
 
 
-class MessageProcessor(object):
+class MessageProcessor:
     def __init__(
         self,
         interpreter: NaturalLanguageInterpreter,
@@ -82,17 +89,23 @@ class MessageProcessor(object):
         """Handle a single message with this processor."""
 
         # preprocess message if necessary
-        tracker = await self.log_message(message)
+        tracker = await self.log_message(message, should_save_tracker=False)
         if not tracker:
             return None
 
         if not self.policy_ensemble or not self.domain:
-            logger.warning(
-                "No policy ensemble or domain set. Skipping action prediction and execution."
+            # save tracker state to continue conversation from this state
+            self._save_tracker(tracker)
+            warnings.warn(
+                "No policy ensemble or domain set. Skipping action prediction "
+                "and execution."
             )
             return None
 
+        start = time.perf_counter()
         await self._predict_and_execute_next_action(message, tracker)
+        duration = time.perf_counter() - start
+        ACTION_PREDICT_EXECUTE_TIME.observe(duration)
         # save tracker state to continue conversation from this state
         self._save_tracker(tracker)
 
@@ -108,8 +121,14 @@ class MessageProcessor(object):
         tracker = self._get_tracker(sender_id)
         if not tracker:
             logger.warning(
-                "Failed to retrieve or create tracker for sender "
-                "'{}'.".format(sender_id)
+                f"Failed to retrieve or create tracker for sender '{sender_id}'."
+            )
+            return None
+
+        if not self.policy_ensemble or not self.domain:
+            # save tracker state to continue conversation from this state
+            warnings.warn(
+                "No policy ensemble or domain set. Skipping action prediction "
             )
             return None
 
@@ -127,7 +146,15 @@ class MessageProcessor(object):
             "tracker": tracker.current_state(EventVerbosity.AFTER_RESTART),
         }
 
-    async def log_message(self, message: UserMessage) -> Optional[DialogueStateTracker]:
+    async def log_message(
+        self, message: UserMessage, should_save_tracker: bool = True
+    ) -> Optional[DialogueStateTracker]:
+        """Log `message` on tracker belonging to the message's conversation_id.
+
+        Optionally save the tracker if `should_save_tracker` is `True`. Tracker saving
+        can be skipped if the tracker returned by this method is used for further
+        processing and saved at a later stage.
+        """
 
         # preprocess message if necessary
         if self.message_preprocessor is not None:
@@ -137,12 +164,14 @@ class MessageProcessor(object):
         tracker = self._get_tracker(message.sender_id)
         if tracker:
             await self._handle_message_with_tracker(message, tracker)
-            # save tracker state to continue conversation from this state
-            self._save_tracker(tracker)
+
+            if should_save_tracker:
+                # save tracker state to continue conversation from this state
+                self._save_tracker(tracker)
         else:
             logger.warning(
                 "Failed to retrieve or create tracker for sender "
-                "'{}'.".format(message.sender_id)
+                f"'{message.sender_id}'."
             )
         return tracker
 
@@ -169,8 +198,7 @@ class MessageProcessor(object):
             self._save_tracker(tracker)
         else:
             logger.warning(
-                "Failed to retrieve or create tracker for sender "
-                "'{}'.".format(sender_id)
+                f"Failed to retrieve or create tracker for sender '{sender_id}'."
             )
         return tracker
 
@@ -236,8 +264,7 @@ class MessageProcessor(object):
 
         if not tracker:
             logger.warning(
-                "Failed to retrieve or create tracker for sender "
-                "'{}'.".format(sender_id)
+                f"Failed to retrieve or create tracker for sender '{sender_id}'."
             )
             return None
 
@@ -270,7 +297,7 @@ class MessageProcessor(object):
     def _log_slots(tracker):
         # Log currently set slots
         slot_values = "\n".join(
-            ["\t{}: {}".format(s.name, s.value) for s in tracker.slots.values()]
+            [f"\t{s.name}: {s.value}" for s in tracker.slots.values()]
         )
         if slot_values.strip():
             logger.debug("[{}] - Current slot values: \n{}".format(tracker.sender_id, slot_values))
@@ -287,19 +314,79 @@ class MessageProcessor(object):
         else:
             return False
 
+    def get_nlu_threshold(self):
+        """Return the NLU confidence threshold"""
+        form_fallback_policy = self.policy_ensemble.look_up_policy(FormFallbackPolicy) if self.policy_ensemble else None
+        if form_fallback_policy:
+            return form_fallback_policy.nlu_threshold or 0.3
+        else:
+            return 0.0
+
+    def get_asr_threshold(self, tracker: DialogueStateTracker):
+        """
+        Return the ASR confidence threshold. Use a default value of 0.4 if a grammar file is used.
+        Otherwise, use the value from the AsrFallbackPolicy.
+        """
+
+        last_bot_utter_template = tracker.latest_bot_utterance.metadata.get("template_name")
+        if last_bot_utter_template:
+            bot_responses = self.domain.templates.get(last_bot_utter_template, [])
+            for bot_response in bot_responses:
+                if bot_response.get("custom") and bot_response.get("custom", {}).get('grammar_file'):
+                    return bot_response.get("custom").get('confidence_threshold', 0.4)
+
+        asr_fallback_policy = self.policy_ensemble.look_up_policy(AsrFallbackPolicy) if self.policy_ensemble else None
+        if asr_fallback_policy:
+            return asr_fallback_policy.nlu_threshold or 0.3
+        else:
+            return 0.0
+
+    def _log_unseen_features(self, parse_data: Dict[Text, Any]) -> None:
+        """Check if the NLU interpreter picks up intents or entities that aren't recognized."""
+
+        domain_is_not_empty = self.domain and not self.domain.is_empty()
+
+        default_intents = [
+            USER_INTENT_RESTART,
+            USER_INTENT_BACK,
+            USER_INTENT_OUT_OF_SCOPE,
+        ]
+
+        intent = parse_data["intent"]["name"]
+        if intent:
+            intent_is_recognized = (
+                domain_is_not_empty and intent in self.domain.intents
+            ) or intent in default_intents
+            if not intent_is_recognized:
+                warnings.warn(
+                    f"Interpreter parsed an intent '{intent}' "
+                    "that is not defined in the domain."
+                )
+
+        entities = parse_data["entities"] or []
+        for element in entities:
+            entity = element["entity"]
+            if entity and domain_is_not_empty and entity not in self.domain.entities:
+                warnings.warn(
+                    f"Interpreter parsed an entity '{entity}' "
+                    "that is not defined in the domain."
+                )
+
     def _get_action(self, action_name):
         return self.domain.action_for_name(action_name, self.action_endpoint)
 
-    async def _parse_message(self, message):
+    async def _parse_message(self, message, tracker: DialogueStateTracker = None):
         # for testing - you can short-cut the NLU part with a message
         # in the format /intent{"entity1": val1, "entity2": val2}
         # parse_data is a dict of intent & entities
         if message.text.startswith(INTENT_MESSAGE_PREFIX):
             parse_data = await RegexInterpreter().parse(
-                message.text, message.message_id
+                message.text, message.message_id, tracker
             )
         else:
-            parse_data = await self.interpreter.parse(message.text, message.message_id)
+            parse_data = await self.interpreter.parse(
+                message.text, message.message_id, tracker
+            )
 
         logger.debug(
             "[{}] - Received user message '{}' with intent '{}' "
@@ -307,6 +394,9 @@ class MessageProcessor(object):
                 message.sender_id, message.text, parse_data["intent"], parse_data["entities"]
             )
         )
+
+        self._log_unseen_features(parse_data)
+
         return parse_data
 
     def create_metadata(
@@ -315,10 +405,18 @@ class MessageProcessor(object):
         requested_slot = tracker.get_slot("requested_slot")
         handoff_active = tracker.get_slot("handoff_active")
 
-        if message.metadata and message.metadata.get("replyType"):
-            metadata = {"type": message.metadata.get("replyType")}
+        metadata = message.metadata or {}
+        if metadata and metadata.get("replyType"):
+            metadata["type"] = metadata.get("replyType")
         else:
-            metadata = {"type": "text"}
+            metadata["type"] = "text"
+
+        if metadata and metadata.get("errorMessage"):
+            metadata["error"] = True
+
+        if metadata and metadata.get("error"):
+            metadata["error"] = True
+            tracker.total_errors += 1
 
         if tracker.active_form.get("name"):
             metadata["active_form"] = tracker.active_form.get("name")
@@ -337,13 +435,19 @@ class MessageProcessor(object):
             elif requested_slot == "id":
                 metadata["type"] = "username"
 
+        if tracker.latest_bot_utterance.metadata and "template_name" in tracker.latest_bot_utterance.metadata:
+            metadata["template_name"] = tracker.latest_bot_utterance.metadata.get("template_name")
+
+        metadata["asr_threshold"] = self.get_asr_threshold(tracker)
+        metadata["nlu_threshold"] = self.get_nlu_threshold()
+
         return metadata
 
-    def log_user_msg(self, message: UserMessage, tracker: DialogueStateTracker):
-        if message.metadata.get('is_pii'):
+    def log_user_msg(self, message, metadata, tracker: DialogueStateTracker):
+        if metadata.get('is_pii'):
             logger.info(f"[{tracker.sender_id}] - User msg: [FILTERED]")
         else:
-            logger.info(f"[{tracker.sender_id}] - User msg: {message.text}")
+            logger.info(f"[{tracker.sender_id}] - User msg: {message}")
 
     async def _handle_message_with_tracker(
         self, message: UserMessage, tracker: DialogueStateTracker
@@ -352,11 +456,20 @@ class MessageProcessor(object):
         if message.parse_data:
             parse_data = message.parse_data
         else:
-            parse_data = await self._parse_message(message)
+            parse_data = await self._parse_message(message, tracker)
+        
+        redacted_text = message.text
+        redacted_entities =  copy.deepcopy(parse_data['entities'])
+        if '/greet' not in redacted_text:
+            redacted_text = UserUttered.redact_msg(redacted_text)
+            if redacted_text != message.text:
+                for e in redacted_entities:
+                    e['value'] = UserUttered.redact_msg(str(e['value']))
 
         missed_intent = not self._is_nlu_threshold_passed(parse_data, tracker)
         message.metadata = self.create_metadata(message, tracker)
-        self.log_user_msg(message, tracker)
+        
+        self.log_user_msg(redacted_text, message.metadata, tracker)
 
         # don't ever directly mutate the tracker
         # - instead pass its events to log
@@ -372,6 +485,8 @@ class MessageProcessor(object):
                 message_id=message.message_id,
                 missed_intent=missed_intent,
                 metadata=message.metadata,
+                text_redacted=redacted_text,
+                entities_redacted=redacted_entities
             ),
             self.domain,
         )
@@ -385,13 +500,13 @@ class MessageProcessor(object):
         )
 
     @staticmethod
-    def _should_handle_message(tracker):
+    def _should_handle_message(tracker: DialogueStateTracker):
         return (
             not tracker.is_paused()
             or tracker.latest_message.intent.get("name") == USER_INTENT_RESTART
         )
 
-    @timeit
+    # @timeit
     async def _predict_and_execute_next_action(
         self, message: UserMessage, tracker: DialogueStateTracker
     ):
@@ -406,7 +521,6 @@ class MessageProcessor(object):
             )
 
         # action loop. predicts actions until we hit action listen
-        start = time.time()
         while (
             should_predict_another_action
             and self._should_handle_message(tracker)
@@ -419,16 +533,14 @@ class MessageProcessor(object):
                 action, tracker, message.output_channel, self.nlg, policy, confidence
             )
             num_predicted_actions += 1
-
-        run_time = round((time.time() - start) * 1000, 2)
-        if run_time > 500:
-            logger.warning(f"Action prediction loop finished in {run_time}ms, num of predicted actions: {num_predicted_actions}")
+            if action and action.name() in ["action_handle_error", ACTION_DEFAULT_FALLBACK_NAME]:
+                tracker.total_errors += 1
 
         if is_action_limit_reached():
             # circuit breaker was tripped
             logger.warning(
                 "Circuit breaker tripped. Stopped predicting "
-                "more actions for sender '{}'".format(tracker.sender_id)
+                f"more actions for sender '{tracker.sender_id}'."
             )
             if self.on_circuit_break:
                 # call a registered callback
@@ -441,7 +553,7 @@ class MessageProcessor(object):
         return not is_listen_action
 
     @staticmethod
-    @timeit
+    # @timeit
     async def _send_bot_messages(
         events: List[Event],
         tracker: DialogueStateTracker,
@@ -454,6 +566,7 @@ class MessageProcessor(object):
 
             if tracker.get_slot("handoff_active") and not tracker.get_slot("initializing_handoff"):
                 e.type_name = "liveAgent"
+            e.metadata["handoff_active"] = tracker.get_slot("handoff_active")
             await output_channel.send_response(tracker.sender_id, e.message())
 
     async def _schedule_reminders(
@@ -503,7 +616,7 @@ class MessageProcessor(object):
                     if j.name == name_to_check:
                         scheduler.remove_job(j.id)
 
-    @timeit
+    # @timeit
     async def _run_action(
         self, action, tracker, output_channel, nlg, policy=None, confidence=None
     ):
@@ -517,12 +630,12 @@ class MessageProcessor(object):
             return self.should_predict_another_action(action.name(), events)
         except Exception as e:
             logger.error(
-                "Encountered an exception while running action '{}'. "
+                "[{}] - Encountered an exception while running action '{}'. "
                 "Bot will continue, but the actions events are lost. "
                 "Please check the logs of your action server for "
-                "more information.".format(action.name())
-            )
-            logger.debug(e, exc_info=True)
+                "more information.".format(tracker.sender_id, action.name())
+            , exc_info=True)
+            # logger.error(e, exc_info=True)
             events = []
 
         self._log_action_on_tracker(tracker, action.name(), events, policy, confidence)
@@ -554,17 +667,16 @@ class MessageProcessor(object):
                     if e.key == "requested_slot" and tracker.active_form:
                         pass
                     else:
-                        logger.warning(
-                            "Action '{0}' set a slot type '{1}' that "
-                            "it never set during the training. This "
-                            "can throw of the prediction. Make sure to "
-                            "include training examples in your stories "
-                            "for the different types of slots this "
-                            "action can return. Remember: you need to "
-                            "set the slots manually in the stories by "
-                            "adding '- slot{{\"{1}\": {2}}}' "
-                            "after the action."
-                            "".format(action_name, e.key, json.dumps(e.value))
+                        warnings.warn(
+                            f"Action '{action_name}' set a slot type '{e.key}' that "
+                            f"it never set during the training. This "
+                            f"can throw of the prediction. Make sure to "
+                            f"include training examples in your stories "
+                            f"for the different types of slots this "
+                            f"action can return. Remember: you need to "
+                            f"set the slots manually in the stories by "
+                            f"adding '- slot{{\"{e.key}\": {e.value}}}' "
+                            f"after the action."
                         )
 
     def _log_action_on_tracker(self, tracker, action_name, events, policy, confidence):
@@ -574,10 +686,9 @@ class MessageProcessor(object):
         if events is None:
             events = []
 
-        start = time.time()
         logger.debug(
             "Action '{}' ended with events '{}'".format(
-                action_name, ["{}".format(e) for e in events]
+                action_name, [f"{e}" for e in events]
             )
         )
 
@@ -594,15 +705,12 @@ class MessageProcessor(object):
             # of the action executed
             e.timestamp = time.time()
             tracker.update(e, self.domain)
-        duration = round((time.time() - start) * 1000, 2)
-        if duration > 500:
-            logger.warning(f"Function _log_action_on_tracker took {duration}ms")
 
     def _get_tracker(self, sender_id: Text) -> Optional[DialogueStateTracker]:
         sender_id = sender_id or UserMessage.DEFAULT_SENDER_ID
         return self.tracker_store.get_or_create_tracker(sender_id)
 
-    def _save_tracker(self, tracker):
+    def _save_tracker(self, tracker: DialogueStateTracker) -> None:
         self.tracker_store.save(tracker)
 
     def _prob_array_for_action(
@@ -630,9 +738,9 @@ class MessageProcessor(object):
                 return result
             else:
                 logger.error(
-                    "Trying to run unknown follow up action '{}'!"
+                    "[{}] - Trying to run unknown follow up action '{}'!"
                     "Instead of running that, we will ignore the action "
-                    "and predict the next action.".format(followup_action)
+                    "and predict the next action.".format(tracker.sender_id, followup_action)
                 )
 
         return self.policy_ensemble.probabilities_using_best_policy(
