@@ -1,13 +1,18 @@
 import asyncio
 import logging
+import warnings
+import multiprocessing
 import os
 import tempfile
 import traceback
 import time
+import multiprocessing
+import typing
 from functools import wraps, reduce
 from inspect import isawaitable
 from typing import Any, Callable, List, Optional, Text, Union
 
+import aiohttp
 from sanic import Sanic, response
 from sanic.request import Request
 from sanic_cors import CORS
@@ -15,37 +20,47 @@ from sanic_jwt import Initialize, exceptions
 
 import dialog_manager
 import rasa
+import rasa.core.brokers.utils
+import rasa.core.utils
 import rasa.utils.common
 import rasa.utils.endpoints
 import rasa.utils.io
-from metrics.constants import CHAT_LATENCY
+from metrics.constants import CHAT_LATENCY, OUTGOING_REQUEST_LATENCY_SEC
+from rasa.core.constants import USER_INTENT_DISCONNECT
 from rasa.core.domain import InvalidDomain
 from rasa.utils.endpoints import EndpointConfig
+
+from rasa import model
 from rasa.constants import (
     MINIMUM_COMPATIBLE_VERSION,
     DEFAULT_MODELS_PATH,
     DEFAULT_DOMAIN_PATH,
     DOCS_BASE_URL,
 )
-from rasa.core import broker
 from rasa.core.agent import load_agent, Agent
 from rasa.core.channels.channel import (
     UserMessage,
     CollectingOutputChannel,
     OutputChannel,
 )
-from rasa.core.events import Event
+from rasa.core.domain import InvalidDomain
+from rasa.core.events import Event, UserUttered, SlotSet
+from rasa.core.lock_store import LockStore
 from rasa.core.test import test
+from rasa.core.tracker_store import TrackerStore
 from rasa.core.trackers import DialogueStateTracker, EventVerbosity
-from rasa.core.utils import dump_obj_as_str_to_file, AvailableEndpoints
-from rasa.model import get_model_subdirectories, fingerprint_from_path
+from rasa.core.utils import AvailableEndpoints
 from rasa.nlu.emulators.no_emulator import NoEmulator
 from rasa.nlu.test import run_evaluation
 from prometheus_client import CollectorRegistry, multiprocess, generate_latest
 from rasa.core.tracker_store import TrackerStore
 import requests
+import aiohttp
 import re
 
+if typing.TYPE_CHECKING:
+    from ssl import SSLContext
+
 logger = logging.getLogger(__name__)
 
 OUTPUT_CHANNEL_QUERY_KEY = "output_channel"
@@ -79,13 +94,22 @@ def _docs(sub_url: Text) -> Text:
     return DOCS_BASE_URL + sub_url
 
 
-def ensure_loaded_agent(app: Sanic):
-    """Wraps a request handler ensuring there is a loaded and usable agent."""
+def ensure_loaded_agent(app: Sanic, require_core_is_ready=False):
+    """Wraps a request handler ensuring there is a loaded and usable agent.
+
+    Require the agent to have a loaded Core model if `require_core_is_ready` is
+    `True`.
+    """
 
     def decorator(f):
         @wraps(f)
         def decorated(*args, **kwargs):
-            if not app.agent or not app.agent.is_ready():
+            # noinspection PyUnresolvedReferences
+            if not app.agent or not (
+                app.agent.is_core_ready()
+                if require_core_is_ready
+                else app.agent.is_ready()
+            ):
                 raise ErrorResponse(
                     409,
                     "Conflict",
@@ -199,7 +223,7 @@ def event_verbosity_parameter(
         )
 
 
-def obtain_tracker_store(agent: "Agent", conversation_id: Text) -> DialogueStateTracker:
+def get_tracker(agent: "Agent", conversation_id: Text) -> DialogueStateTracker:
     tracker = agent.tracker_store.get_or_create_tracker(conversation_id)
     if not tracker:
         raise ErrorResponse(
@@ -224,6 +248,39 @@ async def authenticate(request: Request):
     )
 
 
+def create_ssl_context(
+    ssl_certificate: Optional[Text],
+    ssl_keyfile: Optional[Text],
+    ssl_ca_file: Optional[Text] = None,
+    ssl_password: Optional[Text] = None,
+) -> Optional["SSLContext"]:
+    """Create an SSL context if a proper certificate is passed.
+
+    Args:
+        ssl_certificate: path to the SSL client certificate
+        ssl_keyfile: path to the SSL key file
+        ssl_ca_file: path to the SSL CA file for verification (optional)
+        ssl_password: SSL private key password (optional)
+
+    Returns:
+        SSL context if a valid certificate chain can be loaded, `None` otherwise.
+
+    """
+
+    if ssl_certificate:
+        import ssl
+
+        ssl_context = ssl.create_default_context(
+            purpose=ssl.Purpose.CLIENT_AUTH, cafile=ssl_ca_file
+        )
+        ssl_context.load_cert_chain(
+            ssl_certificate, keyfile=ssl_keyfile, password=ssl_password
+        )
+        return ssl_context
+    else:
+        return None
+
+
 def _create_emulator(mode: Optional[Text]) -> NoEmulator:
     """Create emulator for specified mode.
     If no emulator is specified, we will use the Rasa NLU format."""
@@ -257,6 +314,7 @@ async def _load_agent(
     model_server: Optional[EndpointConfig] = None,
     remote_storage: Optional[Text] = None,
     endpoints: Optional[AvailableEndpoints] = None,
+    lock_store: Optional[LockStore] = None,
 ) -> Agent:
     try:
         tracker_store = None
@@ -264,12 +322,16 @@ async def _load_agent(
         action_endpoint = None
 
         if endpoints:
-            _broker = broker.from_endpoint_config(endpoints.event_broker)
+            _broker = rasa.core.brokers.utils.from_endpoint_config(
+                endpoints.event_broker
+            )
             tracker_store = TrackerStore.find_tracker_store(
                 None, endpoints.tracker_store, _broker
             )
             generator = endpoints.nlg
             action_endpoint = endpoints.action
+            if not lock_store:
+                lock_store = LockStore.find_lock_store(endpoints.lock_store)
 
         loaded_agent = await load_agent(
             model_path,
@@ -277,25 +339,41 @@ async def _load_agent(
             remote_storage,
             generator=generator,
             tracker_store=tracker_store,
+            lock_store=lock_store,
             action_endpoint=action_endpoint,
         )
     except Exception as e:
         logger.debug(traceback.format_exc())
         raise ErrorResponse(
-            500, "LoadingError", "An unexpected error occurred. Error: {}".format(e)
+            500, "LoadingError", f"An unexpected error occurred. Error: {e}"
         )
 
     if not loaded_agent:
         raise ErrorResponse(
             400,
             "BadRequest",
-            "Agent with name '{}' could not be loaded.".format(model_path),
+            f"Agent with name '{model_path}' could not be loaded.",
             {"parameter": "model", "in": "query"},
         )
 
     return loaded_agent
 
 
+def configure_cors(
+    app: Sanic, cors_origins: Union[Text, List[Text], None] = ""
+) -> None:
+    """Configure CORS origins for the given app."""
+
+    # Workaround so that socketio works with requests from other origins.
+    # https://github.com/miguelgrinberg/python-socketio/issues/205#issuecomment-493769183
+    app.config.CORS_AUTOMATIC_OPTIONS = True
+    app.config.CORS_SUPPORTS_CREDENTIALS = True
+
+    CORS(
+        app, resources={r"/*": {"origins": cors_origins or ""}}, automatic_options=True
+    )
+
+
 def add_root_route(app: Sanic):
     @app.get("/")
     async def hello(request: Request):
@@ -305,7 +383,7 @@ def add_root_route(app: Sanic):
 
 def create_app(
     agent: Optional["Agent"] = None,
-    cors_origins: Union[Text, List[Text]] = "*",
+    cors_origins: Union[Text, List[Text], None] = "*",
     auth_token: Optional[Text] = None,
     jwt_secret: Optional[Text] = None,
     jwt_method: Text = "HS256",
@@ -315,10 +393,23 @@ def create_app(
 
     app = Sanic(__name__)
     app.config.RESPONSE_TIMEOUT = 60 * 60
+    app.health_check_urls = []
 
-    # CORS(
-    #     app, resources={r"/*": {"origins": cors_origins or ""}}, automatic_options=True
-    # )
+    if endpoints:
+        if endpoints.duckling:
+            app.health_check_urls.append(endpoints.duckling.url)
+
+        if endpoints.action:
+            action_server_url = endpoints.action.url.replace('/webhook', '')
+            app.health_check_urls.append(f'{action_server_url}/healthz/live')
+
+        if endpoints.idp:
+            app.health_check_urls.append(f'{endpoints.idp.url}/health')
+
+        if endpoints.upm:
+            app.health_check_urls.append(f'{endpoints.upm.url}/liveness')
+
+    configure_cors(app, cors_origins)
 
     # Setup the Sanic-JWT extension
     if jwt_secret and jwt_method:
@@ -335,13 +426,28 @@ def create_app(
         )
 
     app.agent = agent
+    # Initialize shared object of type unsigned int for tracking
+    # the number of active training processes
+    app.active_training_processes = multiprocessing.Value("I", 0)
+
+    async def fetch(session, url, timeout):
+        start = time.perf_counter()
+        result = None
+        try:
+            async with session.get(url, timeout=timeout) as resp:
+                result = resp.status
+        except asyncio.CancelledError:
+            logger.warning(f"Health check task was cancelled. Url: {url}")
+            result = 200
+        except Exception:
+            logger.error(f"Error in fetching data from {url}.", exc_info=True)
+
+        duration = time.perf_counter() - start
+        OUTGOING_REQUEST_LATENCY_SEC.labels(url, result, "GET").observe(duration)
+        return result
 
     @app.listener('after_server_start')
     async def notify_server_started(app, loop):
-        import aiohttp
-        # from config.jaeger_config import initialize_tracer
-        #
-        # initialize_tracer('hydra')
         app.session = aiohttp.ClientSession()
 
 
@@ -366,18 +472,28 @@ def create_app(
             }
         )
 
+    @app.get("/profilerStats")
+    async def profiler_state(request: Request):
+        """ Print the stats of the profiler """
+        app.profiler.stop()
+        return response.html(app.profiler.output_html())
+
+    @app.get("/startProfiler")
+    async def profiler_state(request: Request):
+        """ Print the stats of the profiler """
+        app.profiler.start()
+        return response.text("Profiler started")
+
     @app.get("/healthz/live")
     async def live(request: Request):
         """Check if the server is running and responds with the version."""
-        timeout = 30
-
         try:
             if endpoints.tracker_store:
-                resp = app.agent.tracker_store.test_conn()
+                resp = app.agent.tracker_store.is_healthy()
 
                 if not resp:
-                    logger.error({"event": "HEALTH_CHECK", "message": "Dialog Manager is down."})
-                    raise ErrorResponse(500, "ServerError", "Dialog Manager is not alive")
+                    logger.error({"event": "HEALTH_CHECK", "message": "Tracker store is unhealthy."})
+                    raise ErrorResponse(500, "ServerError", "Tracker store is unhealthy.")
 
             logger.debug({"event": "HEALTH_CHECK", "message": "Dialog Manager is up and running."})
             return response.json({"status": "ok"})
@@ -385,42 +501,72 @@ def create_app(
             logger.error({"event": "HEALTH_CHECK", "message": ex})
             raise ex
         except Exception as ex:
-            traceback_msg = traceback.format_exc().replace("\n", "")
-            logger.error({"event": "HEALTH_CHECK", "message": f"Rasa server is not alive: {traceback_msg}"})
-            raise ErrorResponse(500, "ServerError", f"Dialog Manager is not alive: {traceback_msg}")
+            logger.error({"event": "HEALTH_CHECK", "message": f"Rasa server is not alive"}, exc_info=True)
+            raise ErrorResponse(500, "ServerError", f"Dialog Manager is not alive")
 
     @app.get("/healthz/ready")
     async def ready(request: Request):
         """Check if the server is running and responds with the version."""
         timeout = 30
-        action_resp, duckling_resp = None, None
+        services_status = [False for url in app.health_check_urls]
         _broker = app.agent.tracker_store.event_broker
 
         try:
-            if endpoints.action:
-                action_server_url = endpoints.action.url.replace('/webhook', '')
-                action_resp = await app.session.get(f"{action_server_url}/healthz/live", timeout=timeout)
+            if os.environ.get('READY_CHECK_OFF', None):
+                return response.json({"status": "ok"})
 
-            if endpoints.duckling:
-                duckling_resp = await app.session.get(endpoints.duckling.url, timeout=timeout)
+            start = time.perf_counter()
+            results = await asyncio.gather(*[fetch(app.session, url, timeout) for url in app.health_check_urls])
+            duration = time.perf_counter() - start
+            OUTGOING_REQUEST_LATENCY_SEC.labels("total_healthcheck_duration", "NA", "GET").observe(duration)
+
+            for i in range(len(results)):
+                services_status[i] = results[i]
 
             if _broker and not _broker.is_connected():
                 raise ErrorResponse(500, "ServerError", "Event broker is not available.")
 
-            if not app.agent.is_ready(allow_nlu_only=True):
+            if not app.agent.is_ready():
                 raise ErrorResponse(500, "AgentError", "Agent is not ready yet.")
-            elif action_resp and action_resp.status != 200:
-                raise ErrorResponse(500, "ActionServerError", "Action server is not ready yet.")
-            elif duckling_resp and duckling_resp.status != 200:
-                raise ErrorResponse(500, "DucklingServerError", "Duckling server is not ready yet.")
-
-            logger.debug({"event": "HEALTH_CHECK", "message": "Dialog Manager is ready."})
+            else:
+                for i in range(len(services_status)):
+                    if services_status[i] != 200:
+                        url = app.health_check_urls[i]
+                        raise ErrorResponse(500, "ServerReadiness", f"Failed to get a success response from {url}.")
             return response.json({"status": "ok"})
-        except Exception as ex:
-            traceback_msg = traceback.format_exc().replace("\n", "")
-            err_msg = f'"Dialog Manager is not ready: {traceback_msg}'
-            logger.error({"event": "HEALTH_CHECK", "message": err_msg})
-            raise ErrorResponse(500, "ServerError", err_msg)
+        except asyncio.CancelledError:
+            logger.warning({"event": "HEALTH_CHECK", "message": 'Readiness check task was cancelled.'})
+            return response.json({"status": "ok"})
+        except Exception as err:
+            logger.error({"event": "HEALTH_CHECK", "message": f'Dialog Manager is not ready. {err}'}, exc_info=True)
+            raise ErrorResponse(500, "ServerError", message='Dialog Manager is not ready', details=str(err))
+
+    @app.get("/model/version/<fileName>")
+    async def model_version(request: Request, fileName):
+        try:
+            with open('data/' + fileName) as f:
+                searchTitle = re.search(r'<!--- FileName:\s+(.*?)\s+--->', f.readline())
+                searchDate = re.search(r'<!--- ExportDate:\s+(.*?)\s+--->', f.readline())
+                if (searchTitle and searchTitle.group(1)):
+                    body = {"fileName": searchTitle.group(1), "exportDate": None}
+                    if (searchDate and searchDate.group(1)):
+                        body["exportDate"] = searchDate.group(1)
+                    return response.json(body, status=200)
+                else:
+                    err = {"message": "failed to retrieve version from nlu file"}
+                    return response.json(err, status=500)
+        except Exception as err:
+            logger.error({"event": "MODEL_VERSION", "message": f'File not found. {err}'}, exc_info=True)
+            raise ErrorResponse(500, "ServerError", message=f'file not found: {fileName}', details=str(err))
+
+    @app.route('/model/download/<fileName>')
+    async def download_file(request: Request, fileName):
+        try:
+            return await response.file("data/" + fileName, status=200)
+        except Exception as err:
+            logger.error({"event": "MODEL_VERSION", "message": f'failed to retrieve nlu file. {err}'}, exc_info=True)
+            raise ErrorResponse(500, "ServerError", message=f'failed to retrieve nlu file {fileName}', details=str(err))
+        
 
     # @app.route("/metrics", methods=["GET", "OPTIONS"])
     # async def metrics(request):
@@ -436,8 +582,10 @@ def create_app(
 
         return response.json(
             {
-                "model_file": app.agent.model_directory,
-                "fingerprint": fingerprint_from_path(app.agent.model_directory),
+                "model_file": app.agent.path_to_model_archive
+                or app.agent.model_directory,
+                "fingerprint": model.fingerprint_from_path(app.agent.model_directory),
+                "num_active_training_jobs": app.active_training_processes.value,
             }
         )
 
@@ -446,19 +594,11 @@ def create_app(
     @ensure_loaded_agent(app)
     async def retrieve_tracker(request: Request, conversation_id: Text):
         """Get a dump of a conversation's tracker including its events."""
-        if not app.agent.tracker_store:
-            raise ErrorResponse(
-                409,
-                "Conflict",
-                "No tracker store available. Make sure to "
-                "configure a tracker store when starting "
-                "the server.",
-            )
 
         verbosity = event_verbosity_parameter(request, EventVerbosity.AFTER_RESTART)
         until_time = rasa.utils.endpoints.float_arg(request, "until")
 
-        tracker = obtain_tracker_store(app.agent, conversation_id)
+        tracker = get_tracker(app.agent, conversation_id)
 
         try:
             if until_time is not None:
@@ -469,9 +609,24 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500,
-                "ConversationError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+            )
+
+    @app.get("/conversations/<conversation_id>/tracker/exists")
+    @requires_auth(app, auth_token)
+    @ensure_loaded_agent(app)
+    async def retrieve_tracker(request: Request, conversation_id: Text):
+        """Check if tracker exists in the tracker store with the given conversation_id"""
+        try:
+            tracker = app.agent.tracker_store.retrieve(conversation_id)
+            if tracker:
+                return response.json({"sender_id": conversation_id, "exists": True})
+            else:
+                return response.json({"sender_id": conversation_id, "exists": False})
+        except Exception as e:
+            logger.debug(traceback.format_exc())
+            raise ErrorResponse(
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.post("/conversations/<conversation_id>/tracker/events")
@@ -489,13 +644,17 @@ def create_app(
         if not isinstance(events, list):
             events = [events]
 
+        logger.info(f'[{conversation_id}] - Appending events to the tracker event. Events: {events}')
         events = [Event.from_parameters(event) for event in events]
         events = [event for event in events if event]
 
         if not events:
             logger.warning(
-                "Append event called, but could not extract a valid event. "
-                "Request JSON: {}".format(request.json)
+                "[{}] - Append event called, but could not extract a valid event. "
+                "Request JSON: {}".format(conversation_id, request.json))
+            warnings.warn(
+                f"Append event called, but could not extract a valid event. "
+                f"Request JSON: {request.json}"
             )
             raise ErrorResponse(
                 400,
@@ -505,21 +664,32 @@ def create_app(
             )
 
         verbosity = event_verbosity_parameter(request, EventVerbosity.AFTER_RESTART)
-        tracker = obtain_tracker_store(app.agent, conversation_id)
 
         try:
-            for event in events:
-                tracker.update(event, app.agent.domain)
+            stream_only = False
+            async with app.agent.lock_store.lock(conversation_id):
+                tracker = get_tracker(app.agent, conversation_id)
+                for event in events:
+                    if isinstance(event, UserUttered) and event.intent.get("name") == USER_INTENT_DISCONNECT and not tracker.conversation_ended():
+                        tracker.update(event, app.agent.domain)
+                        if tracker.from_voice_channel():
+                            outcome = 'Hangup'
+                        else:
+                            outcome = 'Disconnected'
+                        val = {"name": "CONVERSATION_END", "message": outcome}
+                        tracker.update(SlotSet("outcome", value=outcome, timestamp=event.timestamp + 1))
+                        tracker.update(SlotSet("system_event", value=val, timestamp=event.timestamp + 2))
+                        stream_only = True
+                    else:
+                        tracker.update(event, app.agent.domain)
 
-            app.agent.tracker_store.save(tracker)
+                app.agent.tracker_store.save(tracker=tracker, timeout=None, stream_only=stream_only)
 
             return response.json(tracker.current_state(verbosity))
         except Exception as e:
-            logger.debug(traceback.format_exc())
+            logger.error(f'[{conversation_id}] - Failed to update events.', exc_info=True)
             raise ErrorResponse(
-                500,
-                "ConversationError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.put("/conversations/<conversation_id>/tracker/events")
@@ -536,19 +706,19 @@ def create_app(
         verbosity = event_verbosity_parameter(request, EventVerbosity.AFTER_RESTART)
 
         try:
-            tracker = DialogueStateTracker.from_dict(
-                conversation_id, request.json, app.agent.domain.slots
-            )
+            async with app.agent.lock_store.lock(conversation_id):
+                tracker = DialogueStateTracker.from_dict(
+                    conversation_id, request.json, app.agent.domain.slots
+                )
+
+                # will override an existing tracker with the same id!
+                app.agent.tracker_store.save(tracker)
 
-            # will override an existing tracker with the same id!
-            app.agent.tracker_store.save(tracker)
             return response.json(tracker.current_state(verbosity))
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500,
-                "ConversationError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.get("/conversations/<conversation_id>/story")
@@ -556,17 +726,9 @@ def create_app(
     @ensure_loaded_agent(app)
     async def retrieve_story(request: Request, conversation_id: Text):
         """Get an end-to-end story corresponding to this conversation."""
-        if not app.agent.tracker_store:
-            raise ErrorResponse(
-                409,
-                "Conflict",
-                "No tracker store available. Make sure to "
-                "configure a tracker store when starting "
-                "the server.",
-            )
 
         # retrieve tracker and set to requested state
-        tracker = obtain_tracker_store(app.agent, conversation_id)
+        tracker = get_tracker(app.agent, conversation_id)
 
         until_time = rasa.utils.endpoints.float_arg(request, "until")
 
@@ -580,9 +742,7 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500,
-                "ConversationError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.post("/conversations/<conversation_id>/execute")
@@ -606,20 +766,24 @@ def create_app(
         verbosity = event_verbosity_parameter(request, EventVerbosity.AFTER_RESTART)
 
         try:
-            tracker = obtain_tracker_store(app.agent, conversation_id)
-            output_channel = _get_output_channel(request, tracker)
-            await app.agent.execute_action(
-                conversation_id, action_to_execute, output_channel, policy, confidence
-            )
+            async with app.agent.lock_store.lock(conversation_id):
+                tracker = get_tracker(app.agent, conversation_id)
+                output_channel = _get_output_channel(request, tracker)
+                await app.agent.execute_action(
+                    conversation_id,
+                    action_to_execute,
+                    output_channel,
+                    policy,
+                    confidence,
+                )
+
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500,
-                "ConversationError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
             )
 
-        tracker = obtain_tracker_store(app.agent, conversation_id)
+        tracker = get_tracker(app.agent, conversation_id)
         state = tracker.current_state(verbosity)
 
         response_body = {"tracker": state}
@@ -643,9 +807,7 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500,
-                "ConversationError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.post("/conversations/<conversation_id>/messages")
@@ -676,16 +838,16 @@ def create_app(
                 {"parameter": "sender", "in": "body"},
             )
 
+        user_message = UserMessage(message, None, conversation_id, parse_data)
+
         try:
-            user_message = UserMessage(message, None, conversation_id, parse_data)
-            tracker = await app.agent.log_message(user_message)
+            async with app.agent.lock_store.lock(conversation_id):
+                tracker = await app.agent.log_message(user_message)
             return response.json(tracker.current_state(verbosity))
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500,
-                "ConversationError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.post("/model/train")
@@ -708,27 +870,36 @@ def create_app(
         temp_dir = tempfile.mkdtemp()
 
         config_path = os.path.join(temp_dir, "config.yml")
-        dump_obj_as_str_to_file(config_path, rjs["config"])
+
+        rasa.utils.io.write_text_file(rjs["config"], config_path)
 
         if "nlu" in rjs:
             nlu_path = os.path.join(temp_dir, "nlu.md")
-            dump_obj_as_str_to_file(nlu_path, rjs["nlu"])
+            rasa.utils.io.write_text_file(rjs["nlu"], nlu_path)
 
         if "stories" in rjs:
             stories_path = os.path.join(temp_dir, "stories.md")
-            dump_obj_as_str_to_file(stories_path, rjs["stories"])
+            rasa.utils.io.write_text_file(rjs["stories"], stories_path)
 
         domain_path = DEFAULT_DOMAIN_PATH
         if "domain" in rjs:
             domain_path = os.path.join(temp_dir, "domain.yml")
-            dump_obj_as_str_to_file(domain_path, rjs["domain"])
+            rasa.utils.io.write_text_file(rjs["domain"], domain_path)
+
+        if rjs.get("save_to_default_model_directory", True) is True:
+            model_output_directory = DEFAULT_MODELS_PATH
+        else:
+            model_output_directory = tempfile.gettempdir()
 
         try:
+            with app.active_training_processes.get_lock():
+                app.active_training_processes.value += 1
+
             model_path = await train_async(
                 domain=domain_path,
                 config=config_path,
                 training_files=temp_dir,
-                output_path=rjs.get("out", DEFAULT_MODELS_PATH),
+                output_path=model_output_directory,
                 force_training=rjs.get("force", False),
             )
 
@@ -741,15 +912,18 @@ def create_app(
             raise ErrorResponse(
                 400,
                 "InvalidDomainError",
-                "Provided domain file is invalid. Error: {}".format(e),
+                f"Provided domain file is invalid. Error: {e}",
             )
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
                 500,
                 "TrainingError",
-                "An unexpected error occurred during training. Error: {}".format(e),
+                f"An unexpected error occurred during training. Error: {e}",
             )
+        finally:
+            with app.active_training_processes.get_lock():
+                app.active_training_processes.value -= 1
 
     def validate_request(rjs):
         if "config" not in rjs:
@@ -780,7 +954,7 @@ def create_app(
 
     @app.post("/model/test/stories")
     @requires_auth(app, auth_token)
-    @ensure_loaded_agent(app)
+    @ensure_loaded_agent(app, require_core_is_ready=True)
     async def evaluate_stories(request: Request):
         """Evaluate stories against the currently loaded model."""
         validate_request_body(
@@ -800,7 +974,7 @@ def create_app(
             raise ErrorResponse(
                 500,
                 "TestingError",
-                "An unexpected error occurred during evaluation. Error: {}".format(e),
+                f"An unexpected error occurred during evaluation. Error: {e}",
             )
 
     @app.post("/model/test/intents")
@@ -831,7 +1005,7 @@ def create_app(
             raise ErrorResponse(409, "Conflict", "Loaded model file not found.")
 
         model_directory = eval_agent.model_directory
-        _, nlu_model = get_model_subdirectories(model_directory)
+        _, nlu_model = model.get_model_subdirectories(model_directory)
 
         try:
             evaluation = await run_evaluation(data_path, nlu_model)
@@ -841,12 +1015,12 @@ def create_app(
             raise ErrorResponse(
                 500,
                 "TestingError",
-                "An unexpected error occurred during evaluation. Error: {}".format(e),
+                f"An unexpected error occurred during evaluation. Error: {e}",
             )
 
     @app.post("/model/predict")
     @requires_auth(app, auth_token)
-    @ensure_loaded_agent(app)
+    @ensure_loaded_agent(app, require_core_is_ready=True)
     async def tracker_predict(request: Request):
         """ Given a list of events, predicts the next action"""
         validate_request_body(
@@ -858,7 +1032,6 @@ def create_app(
         sender_id = UserMessage.DEFAULT_SENDER_ID
         verbosity = event_verbosity_parameter(request, EventVerbosity.AFTER_RESTART)
         request_params = request.json
-
         try:
             tracker = DialogueStateTracker.from_dict(
                 sender_id, request_params, app.agent.domain.slots
@@ -868,7 +1041,7 @@ def create_app(
             raise ErrorResponse(
                 400,
                 "BadRequest",
-                "Supplied events are not valid. {}".format(e),
+                f"Supplied events are not valid. {e}",
                 {"parameter": "", "in": "body"},
             )
 
@@ -893,13 +1066,12 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500,
-                "PredictionError",
-                "An unexpected error occurred. Error: {}".format(e),
+                500, "PredictionError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.post("/model/parse")
     @requires_auth(app, auth_token)
+    @ensure_loaded_agent(app)
     async def parse(request: Request):
         validate_request_body(
             request,
@@ -918,9 +1090,7 @@ def create_app(
             except Exception as e:
                 logger.debug(traceback.format_exc())
                 raise ErrorResponse(
-                    400,
-                    "ParsingError",
-                    "An unexpected error occurred. Error: {}".format(e),
+                    400, "ParsingError", f"An unexpected error occurred. Error: {e}"
                 )
             response_data = emulator.normalise_response_json(parsed_data)
 
@@ -991,7 +1161,7 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ParsingError", "An unexpected error occurred. Error: {}".format(e)
+                500, "ParsingError", f"An unexpected error occurred. Error: {e}"
             )
 
     @app.put("/model")
@@ -1002,6 +1172,7 @@ def create_app(
         model_path = request.json.get("model_file", None)
         model_server = request.json.get("model_server", None)
         remote_storage = request.json.get("remote_storage", None)
+
         if model_server:
             try:
                 model_server = EndpointConfig.from_dict(model_server)
@@ -1010,14 +1181,15 @@ def create_app(
                 raise ErrorResponse(
                     400,
                     "BadRequest",
-                    "Supplied 'model_server' is not valid. Error: {}".format(e),
+                    f"Supplied 'model_server' is not valid. Error: {e}",
                     {"parameter": "model_server", "in": "body"},
                 )
+
         app.agent = await _load_agent(
-            model_path, model_server, remote_storage, endpoints
+            model_path, model_server, remote_storage, endpoints, app.agent.lock_store
         )
 
-        logger.debug("Successfully loaded model '{}'.".format(model_path))
+        logger.debug(f"Successfully loaded model '{model_path}'.")
         return response.json(None, status=204)
 
     @app.delete("/model")
@@ -1025,9 +1197,9 @@ def create_app(
     async def unload_model(request: Request):
         model_file = app.agent.model_directory
 
-        app.agent = Agent()
+        app.agent = Agent(lock_store=app.agent.lock_store)
 
-        logger.debug("Successfully unload model '{}'.".format(model_file))
+        logger.debug(f"Successfully unloaded model '{model_file}'.")
         return response.json(None, status=204)
 
     @app.get("/domain")
