@@ -2,29 +2,21 @@ import asyncio
 from typing import Text
 
 import pytest
-from async_generator import async_generator, yield_
 from sanic import Sanic, response
 
-import rasa.utils.io
 import rasa.core
+import rasa.utils.io
 from rasa.core import jobs, utils
 from rasa.core.agent import Agent, load_agent
 from rasa.core.channels.channel import UserMessage
+from rasa.core.domain import Domain, InvalidDomain
 from rasa.core.interpreter import INTENT_MESSAGE_PREFIX
+from rasa.core.policies.ensemble import PolicyEnsemble
 from rasa.core.policies.memoization import AugmentedMemoizationPolicy
 from rasa.utils.endpoints import EndpointConfig
 from tests.core.conftest import DEFAULT_DOMAIN_PATH_WITH_SLOTS
 
 
-@pytest.fixture(scope="session")
-def loop():
-    loop = asyncio.new_event_loop()
-    asyncio.set_event_loop(loop)
-    loop = rasa.utils.io.enable_async_loop_debugging(loop)
-    yield loop
-    loop.close()
-
-
 def model_server_app(model_path: Text, model_hash: Text = "somehash"):
     app = Sanic(__name__)
     app.number_of_model_requests = 0
@@ -48,15 +40,29 @@ def model_server_app(model_path: Text, model_hash: Text = "somehash"):
 
 
 @pytest.fixture
-@async_generator
 async def model_server(test_server, trained_moodbot_path):
     server = await test_server(
         model_server_app(trained_moodbot_path, model_hash="somehash")
     )
-    await yield_(server)  # python 3.5 compatibility
+    yield server
     await server.close()
 
 
+async def test_training_data_is_reproducible(tmpdir, default_domain):
+    training_data_file = "examples/moodbot/data/stories.md"
+    agent = Agent(
+        "examples/moodbot/domain.yml", policies=[AugmentedMemoizationPolicy()]
+    )
+
+    training_data = await agent.load_data(training_data_file)
+    # make another copy of training data
+    same_training_data = await agent.load_data(training_data_file)
+
+    # test if both datasets are identical (including in the same order)
+    for i, x in enumerate(training_data):
+        assert str(x.as_dialogue()) == str(same_training_data[i].as_dialogue())
+
+
 async def test_agent_train(tmpdir, default_domain):
     training_data_file = "examples/moodbot/data/stories.md"
     agent = Agent(
@@ -64,6 +70,7 @@ async def test_agent_train(tmpdir, default_domain):
     )
 
     training_data = await agent.load_data(training_data_file)
+
     agent.train(training_data)
     agent.persist(tmpdir.strpath)
 
@@ -156,7 +163,7 @@ async def test_agent_with_model_server_in_thread(
         agent, model_server=model_endpoint_config
     )
 
-    await asyncio.sleep(3)
+    await asyncio.sleep(5)
 
     assert agent.fingerprint == "somehash"
     assert hash(agent.domain) == hash(moodbot_domain)
@@ -171,7 +178,6 @@ async def test_agent_with_model_server_in_thread(
 
 
 async def test_wait_time_between_pulls_without_interval(model_server, monkeypatch):
-
     monkeypatch.setattr(
         "rasa.core.agent.schedule_model_pulling", lambda *args: 1 / 0
     )  # will raise an exception
@@ -194,6 +200,53 @@ async def test_load_agent(trained_model):
     assert agent.model_directory is not None
 
 
+@pytest.mark.parametrize(
+    "domain, policy_config",
+    [({"forms": ["restaurant_form"]}, {"policies": [{"name": "MemoizationPolicy"}]})],
+)
+def test_form_without_form_policy(domain, policy_config):
+    with pytest.raises(InvalidDomain) as execinfo:
+        Agent(
+            domain=Domain.from_dict(domain),
+            policies=PolicyEnsemble.from_dict(policy_config),
+        )
+    assert "haven't added the FormPolicy" in str(execinfo.value)
+
+
+@pytest.mark.parametrize(
+    "domain, policy_config",
+    [
+        (
+            {
+                "intents": [{"affirm": {"triggers": "utter_ask_num_people"}}],
+                "actions": ["utter_ask_num_people"],
+            },
+            {"policies": [{"name": "MemoizationPolicy"}]},
+        )
+    ],
+)
+def test_trigger_without_mapping_policy(domain, policy_config):
+    with pytest.raises(InvalidDomain) as execinfo:
+        Agent(
+            domain=Domain.from_dict(domain),
+            policies=PolicyEnsemble.from_dict(policy_config),
+        )
+    assert "haven't added the MappingPolicy" in str(execinfo.value)
+
+
+@pytest.mark.parametrize(
+    "domain, policy_config",
+    [({"intents": ["affirm"]}, {"policies": [{"name": "TwoStageFallbackPolicy"}]})],
+)
+def test_two_stage_fallback_without_deny_suggestion(domain, policy_config):
+    with pytest.raises(InvalidDomain) as execinfo:
+        Agent(
+            domain=Domain.from_dict(domain),
+            policies=PolicyEnsemble.from_dict(policy_config),
+        )
+    assert "The intent 'out_of_scope' must be present" in str(execinfo.value)
+
+
 async def test_agent_update_model_none_domain(trained_model):
     agent = await load_agent(model_path=trained_model)
     agent.update_model(
@@ -203,7 +256,7 @@ async def test_agent_update_model_none_domain(trained_model):
     sender_id = "test_sender_id"
     message = UserMessage("hello", sender_id=sender_id)
     await agent.handle_message(message)
-    tracker = await agent.tracker_store.get_or_create_tracker(sender_id)
+    tracker = agent.tracker_store.get_or_create_tracker(sender_id)
 
     # UserUttered event was added to tracker, with correct intent data
     assert tracker.events[1].intent["name"] == "greet"
