@@ -1,3 +1,4 @@
+import aiohttp
 import asyncio
 import datetime
 import uuid
@@ -5,8 +6,11 @@ import uuid
 import pytest
 from aioresponses import aioresponses
 
+from unittest.mock import patch
+
 import rasa.utils.io
 from rasa.core import jobs
+from rasa.core.agent import Agent
 from rasa.core.channels.channel import CollectingOutputChannel, UserMessage
 from rasa.core.events import (
     ActionExecuted,
@@ -16,19 +20,20 @@ from rasa.core.events import (
     Restarted,
     UserUttered,
 )
+from rasa.core.trackers import DialogueStateTracker
+from rasa.core.slots import Slot
+from rasa.core.processor import MessageProcessor
 from rasa.core.interpreter import RasaNLUHttpInterpreter
 from rasa.core.processor import MessageProcessor
 from rasa.utils.endpoints import EndpointConfig
 from tests.utilities import json_of_latest_request, latest_request
 
+from tests.core.conftest import DEFAULT_DOMAIN_PATH_WITH_SLOTS
+from rasa.core.domain import Domain
 
-@pytest.fixture(scope="session")
-def loop():
-    loop = asyncio.new_event_loop()
-    asyncio.set_event_loop(loop)
-    loop = rasa.utils.io.enable_async_loop_debugging(loop)
-    yield loop
-    loop.close()
+import logging
+
+logger = logging.getLogger(__name__)
 
 
 async def test_message_processor(
@@ -62,6 +67,30 @@ async def test_parsing(default_processor: MessageProcessor):
     assert parsed["entities"][0]["entity"] == "name"
 
 
+async def test_log_unseen_feature(default_processor: MessageProcessor):
+    message = UserMessage('/dislike{"test_entity": "RASA"}')
+    parsed = await default_processor._parse_message(message)
+    with pytest.warns(UserWarning) as record:
+        default_processor._log_unseen_features(parsed)
+    assert len(record) == 2
+    assert (
+        record[0].message.args[0]
+        == "Interpreter parsed an intent 'dislike' that is not defined in the domain."
+    )
+    assert (
+        record[1].message.args[0]
+        == "Interpreter parsed an entity 'test_entity' that is not defined in the domain."
+    )
+
+
+async def test_default_intent_recognized(default_processor: MessageProcessor):
+    message = UserMessage("/restart")
+    parsed = await default_processor._parse_message(message)
+    with pytest.warns(None) as record:
+        default_processor._log_unseen_features(parsed)
+    assert len(record) == 0
+
+
 async def test_http_parsing():
     message = UserMessage("lunch?")
 
@@ -82,25 +111,56 @@ async def test_http_parsing():
         assert r
 
 
+async def mocked_parse(self, text, message_id=None, tracker=None):
+    """Mock parsing a text message and augment it with the slot
+    value from the tracker's state."""
+
+    return {
+        "intent": {"name": "", "confidence": 0.0},
+        "entities": [],
+        "text": text,
+        "requested_language": tracker.get_slot("requested_language"),
+    }
+
+
+async def test_parsing_with_tracker():
+    tracker = DialogueStateTracker.from_dict("1", [], [Slot("requested_language")])
+
+    # we'll expect this value 'en' to be part of the result from the interpreter
+    tracker._set_slot("requested_language", "en")
+
+    endpoint = EndpointConfig("https://interpreter.com")
+    with aioresponses() as mocked:
+        mocked.post("https://interpreter.com/parse", repeat=True, status=200)
+
+        # mock the parse function with the one defined for this test
+        with patch.object(RasaNLUHttpInterpreter, "parse", mocked_parse):
+            interpreter = RasaNLUHttpInterpreter(endpoint=endpoint)
+            agent = Agent(None, None, interpreter)
+            result = await agent.parse_message_using_nlu_interpreter("lunch?", tracker)
+
+            assert result["requested_language"] == "en"
+
+
 async def test_reminder_scheduled(
     default_channel: CollectingOutputChannel, default_processor: MessageProcessor
 ):
     sender_id = uuid.uuid4().hex
 
     reminder = ReminderScheduled("utter_greet", datetime.datetime.now())
-    tracker = await default_processor.tracker_store.get_or_create_tracker(sender_id)
+    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)
 
     tracker.update(UserUttered("test"))
     tracker.update(ActionExecuted("action_reminder_reminder"))
     tracker.update(reminder)
 
-    await default_processor.tracker_store.save(tracker)
+    default_processor.tracker_store.save(tracker)
     await default_processor.handle_reminder(
         reminder, sender_id, default_channel, default_processor.nlg
     )
 
     # retrieve the updated tracker
-    t = await default_processor.tracker_store.retrieve(sender_id)
+    t = default_processor.tracker_store.retrieve(sender_id)
     assert t.events[-4] == UserUttered(None)
     assert t.events[-3] == ActionExecuted("utter_greet")
     assert t.events[-2] == BotUttered(
@@ -113,9 +173,33 @@ async def test_reminder_scheduled(
             "image": None,
             "custom": None,
         },
+        metadata={"template_name": "utter_greet"}
     )
     assert t.events[-1] == ActionExecuted("action_listen")
 
+async def test_redaction(default_processor: MessageProcessor):
+    message_1 = UserUttered.redact_msg("hello")
+    message_2 = UserUttered.redact_msg("7/12/1996")
+    message_3 = UserUttered.redact_msg("five dollars")
+    message_4 = UserUttered.redact_msg("5 dollars")
+    message_5 = UserUttered.redact_msg("5.50")
+    message_6 = UserUttered.redact_msg("1a2a3")
+    message_7 = UserUttered.redact_msg("one hundred")
+    message_8 = UserUttered.redact_msg("5 and eight")
+    message_9 = UserUttered.redact_msg("one thousand six hundred sixty six" )
+    message_10 = UserUttered.redact_msg("one thousand and six hundred and sixty six")
+    assert message_1 == "hello"
+    assert message_2 == "****"
+    assert message_3 == "**** dollars"
+    assert message_4 == "**** dollars"
+    assert message_5 == "****"
+    assert message_6 == "****"
+    assert message_7 == "****"
+    assert message_8 == "**** and ****"
+    assert message_9 == "****"
+    assert message_10 == "****"
+
+
 
 async def test_reminder_aborted(
     default_channel: CollectingOutputChannel, default_processor: MessageProcessor
@@ -125,18 +209,18 @@ async def test_reminder_aborted(
     reminder = ReminderScheduled(
         "utter_greet", datetime.datetime.now(), kill_on_user_message=True
     )
-    tracker = await default_processor.tracker_store.get_or_create_tracker(sender_id)
+    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)
 
     tracker.update(reminder)
     tracker.update(UserUttered("test"))  # cancels the reminder
 
-    await default_processor.tracker_store.save(tracker)
+    default_processor.tracker_store.save(tracker)
     await default_processor.handle_reminder(
         reminder, sender_id, default_channel, default_processor.nlg
     )
 
     # retrieve the updated tracker
-    t = await default_processor.tracker_store.retrieve(sender_id)
+    t = default_processor.tracker_store.retrieve(sender_id)
     assert len(t.events) == 3  # nothing should have been executed
 
 
@@ -146,7 +230,7 @@ async def test_reminder_cancelled(
     sender_ids = [uuid.uuid4().hex, uuid.uuid4().hex]
     trackers = []
     for sender_id in sender_ids:
-        tracker = await default_processor.tracker_store.get_or_create_tracker(sender_id)
+        tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)
 
         tracker.update(UserUttered("test"))
         tracker.update(ActionExecuted("action_reminder_reminder"))
@@ -161,7 +245,7 @@ async def test_reminder_cancelled(
     trackers[0].update(ReminderCancelled("utter_greet"))
 
     for tracker in trackers:
-        await default_processor.tracker_store.save(tracker)
+        default_processor.tracker_store.save(tracker)
         await default_processor._schedule_reminders(
             tracker.events, tracker, default_channel, default_processor.nlg
         )
@@ -174,13 +258,13 @@ async def test_reminder_cancelled(
     assert len((await jobs.scheduler()).get_jobs()) == 1
 
     # execute the jobs
-    await asyncio.sleep(3)
+    await asyncio.sleep(5)
 
-    tracker_0 = await default_processor.tracker_store.retrieve(sender_ids[0])
+    tracker_0 = default_processor.tracker_store.retrieve(sender_ids[0])
     # there should be no utter_greet action
     assert ActionExecuted("utter_greet") not in tracker_0.events
 
-    tracker_1 = await default_processor.tracker_store.retrieve(sender_ids[1])
+    tracker_1 = default_processor.tracker_store.retrieve(sender_ids[1])
     # there should be utter_greet action
     assert ActionExecuted("utter_greet") in tracker_1.events
 
@@ -193,17 +277,17 @@ async def test_reminder_restart(
     reminder = ReminderScheduled(
         "utter_greet", datetime.datetime.now(), kill_on_user_message=False
     )
-    tracker = await default_processor.tracker_store.get_or_create_tracker(sender_id)
+    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)
 
     tracker.update(reminder)
     tracker.update(Restarted())  # cancels the reminder
     tracker.update(UserUttered("test"))
 
-    await default_processor.tracker_store.save(tracker)
+    default_processor.tracker_store.save(tracker)
     await default_processor.handle_reminder(
         reminder, sender_id, default_channel, default_processor.nlg
     )
 
     # retrieve the updated tracker
-    t = await default_processor.tracker_store.retrieve(sender_id)
+    t = default_processor.tracker_store.retrieve(sender_id)
     assert len(t.events) == 4  # nothing should have been executed
